
#include "ruby.h"

#include "lsapilib.h"
#include <errno.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/wait.h>

//RUBY_EXTERN VALUE ruby_errinfo;
RUBY_EXTERN VALUE rb_stdin;
RUBY_EXTERN VALUE rb_stdout;
#if RUBY_VERSION_CODE < 180
RUBY_EXTERN VALUE rb_defout;
#endif

static VALUE orig_stdin;
static VALUE orig_stdout;
#if RUBY_VERSION_CODE < 180
static VALUE orig_defout;
#endif

static VALUE orig_env;

static VALUE env_copy;

static VALUE lsapi_env;


static VALUE lsapi_objrefs;

typedef struct lsapi_data
{
    LSAPI_Request * req;
    VALUE           env;
}lsapi_data;


static VALUE cLSAPI;

static VALUE s_req = Qnil;
static lsapi_data * s_req_data;

static int  s_children      = 0;
static int  s_cur_children  = 0;
static int  s_req_processed = 0;
static int  s_max_reqs      = 1000000;
static int  s_max_idle_secs = 60;
static int  s_listenFd      = -1;
static int  s_ppid          = 0;
static int  s_max_grp_idle_secs = 0;

static void lsapi_ruby_setenv(const char *name, const char *value)
{
    if (!name) return;

    if (value && *value)
	    ruby_setenv(name, value);
    else
        ruby_unsetenv(name);
}



static void lsapi_mark( lsapi_data * data )
{
    rb_gc_mark( data->env );
}

static void lsapi_free_data( lsapi_data * data )
{
    free( data );
}

static int add_env_rails( const char * pKey, int keyLen, const char * pValue, int valLen,
                         void * arg )
{
    char * p;
    int len;
    //Fixup some environment variables for rails
    switch( *pKey )
    {
    case 'Q':
        if ( strcmp( pKey, "QUERY_STRING" ) == 0 )
        {
            if ( !*pValue )
                return 1;
        }
        break;
    case 'R':
        if (( *(pKey+8) == 'U' )&&( strcmp( pKey, "REQUEST_URI" ) == 0 ))
        {
            p = strchr( pValue, '?' );
            if ( p )
            {
                len = valLen - ( p - pValue ) - 1;
                valLen = p - pValue;
                *p++ = 0;
            }
            else
            {
                p = "";
                len = 0;
            }
            rb_hash_aset( lsapi_env,rb_tainted_str_new("QUERY_STRING", 12),
                        rb_tainted_str_new(p, len));
            rb_hash_aset( lsapi_env,rb_tainted_str_new("PATH_INFO", 9),
                        rb_tainted_str_new(pValue, valLen));
        }
        break;
    case 'S':
        if ( strcmp( pKey, "SCRIPT_NAME" ) == 0 )
        {
            pValue = "/";
            valLen = 1;
        }
        break;
    case 'P':
        if ( strcmp( pKey, "PATH_INFO" ) == 0 )
            return 1;
    default:
        break;
    }
    
    //lsapi_ruby_setenv(pKey, pValue );
    
    rb_hash_aset( lsapi_env,rb_tainted_str_new(pKey, keyLen),
                    rb_tainted_str_new(pValue, valLen));
    return 1;
}

static int add_env_no_fix( const char * pKey, int keyLen, const char * pValue, int valLen,
                         void * arg )
{
    rb_hash_aset( lsapi_env,rb_tainted_str_new(pKey, keyLen),
                    rb_tainted_str_new(pValue, valLen));
    return 1;
}

typedef int (*fn_add_env)( const char * pKey, int keyLen, const char * pValue, int valLen,
                         void * arg );
                         
fn_add_env s_fn_add_env = add_env_no_fix;
                         
static void clear_env()
{
    //rb_funcall( lsapi_env, rb_intern( "clear" ), 0 );
    rb_funcall( lsapi_env, rb_intern( "replace" ), 1, env_copy );
}

static void setup_cgi_env( lsapi_data * data )
{
    clear_env();

    LSAPI_ForeachHeader_r( data->req, s_fn_add_env, data );
    LSAPI_ForeachEnv_r( data->req, s_fn_add_env, data );
    
}



static int s_stop;
void lsapi_cleanup(int signal)
{
    s_stop = signal;
}

void lsapi_sigchild( int signal )
{
    int status, pid;
    while( 1 )
    {
        pid = waitpid( -1, &status, WNOHANG|WUNTRACED );
        if ( pid <= 0 )
        {
            break;
        }
        --s_cur_children;
    }

}

static int lsapi_ruby_accept( int fdListen )
{
    int         fd;
    int         nodelay = 1;
    socklen_t   len;
    char        achPeer[128];

    len = sizeof( achPeer );
    fd = accept( s_listenFd, (struct sockaddr *)&achPeer, &len );
    if ( fd != -1 )
    {
        if (((struct sockaddr *)&achPeer)->sa_family == AF_INET )
        {
            setsockopt( fd, IPPROTO_TCP, TCP_NODELAY,
                    (char *)&nodelay, sizeof(nodelay));
        }
    }
    return fd;
    
}

static int lsapi_fork_child()
{
    struct sigaction act, old_term, old_quit, old_int,
                    old_usr1, old_child;

    int             wait_secs = 0;                    
    int             ret;
    fd_set          readfds;
    struct timeval  timeout;

    setsid();

    act.sa_flags = 0;
    act.sa_handler = lsapi_sigchild;
    if( sigaction( SIGCHLD, &act, &old_child ) )
    {
        perror( "Can't set signal handler for SIGCHILD" );
        return -1;
    }

    /* Set up handler to kill children upon exit */
    act.sa_flags = 0;
    act.sa_handler = lsapi_cleanup;
    if( sigaction( SIGTERM, &act, &old_term ) ||
        sigaction( SIGINT,  &act, &old_int  ) ||
        sigaction( SIGUSR1, &act, &old_usr1 ) ||
        sigaction( SIGQUIT, &act, &old_quit ))
    {
        perror( "Can't set signals" );
        return -1;
    }
    s_stop = 0;
    while( !s_stop )
    {
        if ( s_cur_children >= (s_children << 1 ) )
        {
            usleep( 10000 );
            continue;
        }
        
        FD_ZERO( &readfds );
        FD_SET( s_listenFd, &readfds );
        timeout.tv_sec = 1; timeout.tv_usec = 0;
        if ((ret = rb_thread_select(s_listenFd+1, &readfds, NULL, NULL, &timeout)) == 1 )
        {
            if ( s_cur_children >= 0 )
            {
                usleep( 10 );
                FD_ZERO( &readfds );
                FD_SET( s_listenFd, &readfds );
                timeout.tv_sec = 0; timeout.tv_usec = 0;
                if ( select(s_listenFd+1, &readfds, NULL, NULL, &timeout) == 0 )
                    continue;
            }
        }
        else if ( ret == -1 )
            break;
        else
        {
            if (s_ppid && (kill(s_ppid, 0) == -1)&&(errno == ESRCH))
                break;
            if (s_max_grp_idle_secs)
            {
                if ( s_cur_children <= 0 )
                {
                    ++wait_secs;
                    if ( wait_secs > s_max_grp_idle_secs )
                        return -1;
                }
                else
                    wait_secs = 0;
            }
            continue;
        }

        g_req.m_fd = lsapi_ruby_accept( s_listenFd );
        if ( g_req.m_fd != -1 )
        {
            int pid = fork();
            if ( !pid )
            {
                s_children = 0;
                s_ppid = getppid();
                
                /* don't catch our signals */
                sigaction( SIGCHLD, &old_child, 0 );
                sigaction( SIGTERM, &old_term, 0 );
                sigaction( SIGQUIT, &old_quit, 0 );
                sigaction( SIGINT,  &old_int,  0 );
                sigaction( SIGUSR1, &old_usr1, 0 );
                return 0;
            }
            else if ( pid == -1 )
            {
                perror( "fork() failed, please increase process limit" );
            }
            else
            {
                ++s_cur_children;
            }
            close( g_req.m_fd );
            g_req.m_fd = -1;
            
        }
        else
        {
            if (( errno == EINTR )||( errno == EAGAIN))
                continue;
            perror( "accept() failed" );
            return -1;
        }
    }
    kill( -getpgrp(), SIGUSR1 );
    return -1;

}

static VALUE lsapi_s_accept( VALUE self )
{
    int             fd;
    int             ret;
    int             wait_secs;
    fd_set          readfds;
    struct timeval  timeout;

    LSAPI_Finish_r( &g_req );
    if ( s_children > 1)
    {
        if ( s_listenFd != -1 )
            if ( lsapi_fork_child() == -1 )
                return Qnil;
    }
    if ( s_req_processed >= s_max_reqs )
        return Qnil;
    while( 1 )
    {
        if ( g_req.m_fd != -1 )
        {
            fd = g_req.m_fd;
        }
        else if ( s_listenFd != -1 )
            fd = s_listenFd;
        else
            return Qnil;
        wait_secs = 0;
        while( LSAPI_IsRunning() )
        {
            FD_ZERO( &readfds );
            FD_SET( fd, &readfds );
            timeout.tv_sec = 1;
            timeout.tv_usec = 0;
            if (rb_thread_select(fd+1, &readfds, NULL, NULL, &timeout) < 1)
            {
                ++wait_secs;
                if (( s_max_idle_secs > 0 )&&(wait_secs >= s_max_idle_secs ))
                    return Qnil;
                if ( s_ppid &&(kill(s_ppid, 0) == -1)&&(errno == ESRCH))
                    return Qnil;
            }
            else if ( fd == s_listenFd )
            {
                g_req.m_fd = lsapi_ruby_accept( s_listenFd );
                if ( g_req.m_fd != -1 )
                {
                    break;
                }
            }
            else
                break;
        }
    
        ret = LSAPI_Accept_r( &g_req );
    
        if ( !ret )
        {
            ++s_req_processed;
            setup_cgi_env( s_req_data );
            return s_req;
        }
        else if (( g_req.m_fd == -1 )&&( fd != s_listenFd ))
            continue;
        else
            break;
    }
    return Qnil;
}


static int chdir_file( const char * pFile )
{
    char * p = strrchr( pFile, '/' );
    int ret;
    if ( !p )
        return -1;
    *p = 0;
    ret = chdir( pFile );
    *p = '/';
    return ret;
}

static VALUE lsapi_eval_string_wrap(VALUE self, VALUE str)
{
    if (rb_safe_level() >= 4)
    {
        Check_Type(str, T_STRING);
    }
    else
    {
        Check_SafeStr(str);
    }
    return rb_eval_string_wrap(StringValuePtr(str), NULL);
}

static VALUE lsapi_process( VALUE self )
{
    lsapi_data *data;
    const char * pScriptPath;
    Data_Get_Struct(self,lsapi_data, data);
    pScriptPath = LSAPI_GetScriptFileName_r( data->req );
//    if ( chdir_file( pScriptPath ) == -1 )
//    {
//        lsapi_send_error( 404 );
//    }
//    rb_load_file( pScriptPath );
    return Qnil;
}


static VALUE lsapi_putc(VALUE self, VALUE c)
{
    char ch = NUM2CHR(c);
//    lsapi_data *data;
//    Data_Get_Struct(self,lsapi_data, data);
    if ( LSAPI_Write_r( &g_req, &ch, 1 ) == 1 )
        return c;
    else
        return INT2NUM( EOF );
}


static VALUE lsapi_write( VALUE self, VALUE str )
{
    lsapi_data *data;
    int len;
//    Data_Get_Struct(self,lsapi_data, data);
//    len = LSAPI_Write_r( data->req, RSTRING(str)->ptr, RSTRING(str)->len );
    if (TYPE(str) != T_STRING)
	    str = rb_obj_as_string(str);
    len = LSAPI_Write_r( &g_req, RSTRING(str)->ptr, RSTRING(str)->len );
    return INT2NUM( len );
}

static VALUE lsapi_print( int argc, VALUE *argv, VALUE out )
{
    int i;
    VALUE line;

    /* if no argument given, print `$_' */
    if (argc == 0)
    {
      argc = 1;
      line = rb_lastline_get();
      argv = &line;
    }
    for (i = 0; i<argc; i++)
    {
        if (!NIL_P(rb_output_fs) && i>0)
        {
            lsapi_write(out, rb_output_fs);
        }
        switch (TYPE(argv[i]))
        {
        case T_NIL:
            lsapi_write(out, rb_str_new2("nil"));
            break;
        default:
            lsapi_write(out, argv[i]);
            break;
        }
    }
    if (!NIL_P(rb_output_rs))
    {
        lsapi_write(out, rb_output_rs);
    }

    return Qnil;
}

static VALUE lsapi_printf(int argc, VALUE *argv, VALUE out)
{
    lsapi_write(out, rb_f_sprintf(argc, argv));
    return Qnil;
}

static VALUE lsapi_puts _((int, VALUE*, VALUE));

static VALUE lsapi_puts_ary(VALUE ary, VALUE out)
{
    VALUE tmp;
    int i;

    for (i=0; i<RARRAY(ary)->len; i++)
    {
        tmp = RARRAY(ary)->ptr[i];
	    if (rb_inspecting_p(tmp))
        {
	        tmp = rb_str_new2("[...]");
	    }
	    lsapi_puts(1, &tmp, out);
    }
    return Qnil;
}

static VALUE lsapi_puts(int argc, VALUE *argv, VALUE out)
{
    int i;
    VALUE line;

    /* if no argument given, print newline. */
    if (argc == 0)
    {
        lsapi_write(out, rb_default_rs);
        return Qnil;
    }
    for (i=0; i<argc; i++)
    {
        switch (TYPE(argv[i]))
        {
        case T_NIL:
            line = rb_str_new2("nil");
            break;
        case T_ARRAY:
            rb_protect_inspect(lsapi_puts_ary, argv[i], out);
            continue;
        default:
            line = argv[i];
            break;
        }
        line = rb_obj_as_string(line);
        lsapi_write(out, line);
        if (RSTRING(line)->ptr[RSTRING(line)->len-1] != '\n')
        {
            lsapi_write(out, rb_default_rs);
        }
    }

    return Qnil;
}


static VALUE lsapi_addstr(VALUE out, VALUE str)
{
    lsapi_write(out, str);
    return out;
}

static VALUE lsapi_flush( VALUE self )
{
    //lsapi_data *data;
    //Data_Get_Struct(self,lsapi_data, data);
    LSAPI_Flush_r( &g_req ); 
    return Qnil;
}

static VALUE lsapi_getc( VALUE self )
{
    int ch;
    //lsapi_data *data;
    //Data_Get_Struct(self,lsapi_data, data);
    if (rb_safe_level() >= 4 && !OBJ_TAINTED(self))
    {
        rb_raise(rb_eSecurityError, "Insecure: operation on untainted IO");
    }
    ch = LSAPI_ReqBodyGetChar_r( &g_req );
    return INT2NUM( ch );
}

static VALUE lsapi_gets( VALUE self )
{
    VALUE str;
    int getLF = 0;
    char buff[4096];
    int len;
    if (rb_safe_level() >= 4 && !OBJ_TAINTED(self))
    {
        rb_raise(rb_eSecurityError, "Insecure: operation on untainted IO");
    }
    str = rb_str_buf_new( 1024 );
    OBJ_TAINT(str);
    while( !getLF )
    {
        len = LSAPI_ReqBodyGetLine_r( &g_req, buff, 4096, &getLF );
        if ( len > 0 )
            rb_str_buf_cat( str, buff, len );
        
    }
    return str;
}


static VALUE lsapi_read(int argc, VALUE *argv, VALUE self)
{
    VALUE str;
    lsapi_data *data;
    int n;
    int remain;
    char buff[8192];

    if (rb_safe_level() >= 4 && !OBJ_TAINTED(self))
    {
        rb_raise(rb_eSecurityError, "Insecure: operation on untainted IO");
    }

    Data_Get_Struct(self, lsapi_data, data);

    remain = LSAPI_GetReqBodyRemain();
    if ( remain <= 0 )
        return Qnil;
    if (argc != 0)
    {
        n = NUM2INT(argv[0]);
        if ( remain > n )
            remain = n;
    }
    str = rb_str_buf_new( remain );
    OBJ_TAINT(str);
    while( remain > 0 )
    {
        n = LSAPI_ReadReqBody_r(data->req, buff,
                        (remain > 8192)?8192:remain );
        if ( n > 0 )
        {
            rb_str_buf_cat( str, buff, n );
            remain -= n;
        }
        else if ( n <= 0 )
        {
            //FIXME: broken connection
            break;
        }
    }
    return str;
}

static VALUE lsapi_eof(VALUE self)
{
//    lsapi_data *data;
//
//    if (rb_safe_level() >= 4 && !OBJ_TAINTED(self))
//    {
//        rb_raise(rb_eSecurityError, "Insecure: operation on untainted IO");
//    }
//    Data_Get_Struct(self, lsapi_data, data);
    return (LSAPI_GetReqBodyRemain( &g_req ) <= 0) ? Qtrue : Qfalse;
}

static VALUE lsapi_binmode(VALUE self)
{
//    if (rb_safe_level() >= 4 && !OBJ_TAINTED(self)) {
//        rb_raise(rb_eSecurityError, "Insecure: operation on untainted IO");
//    }
    return self;
}

static VALUE lsapi_isatty(VALUE self)
{
//    if (rb_safe_level() >= 4 && !OBJ_TAINTED(self)) {
//        rb_raise(rb_eSecurityError, "Insecure: operation on untainted IO");
//    }
    return Qfalse;
}

static VALUE lsapi_sync(VALUE self)
{
//    if (rb_safe_level() >= 4 && !OBJ_TAINTED(self)) {
//        rb_raise(rb_eSecurityError, "Insecure: operation on untainted IO");
//    }
    return Qfalse;
}

static VALUE lsapi_setsync(VALUE self,VALUE sync)
{
//    if (rb_safe_level() >= 4 && !OBJ_TAINTED(self)) {
//        rb_raise(rb_eSecurityError, "Insecure: operation on untainted IO");
//    }
    return Qfalse;
}

void Init_lsapi()
{
    VALUE remove_env;
    int n;
    char * p = getenv( "LSAPI_CHILDREN" );
    if ( p )
    {
        s_children = atoi( p );
        ruby_unsetenv( "LSAPI_CHILDREN" );
    }
    
    p = getenv( "LSAPI_MAX_REQS" );
    if ( p )
    {
        n = atoi( p );
        if ( n > 0 )
            s_max_reqs = n;
        ruby_unsetenv( "LSAPI_MAX_REQS" );
    }
    
    p = getenv( "LSAPI_MAX_IDLE" );
    if ( p )
    {
        n = atoi( p );
        s_max_idle_secs = n;
        ruby_unsetenv( "LSAPI_MAX_IDLE" );
    }

    p = getenv( "LSAPI_PGRP_MAX_IDLE" );
    if ( p )
    {
        s_max_grp_idle_secs = atoi( p );
        ruby_unsetenv( "LSAPI_PGRP_MAX_IDLE" );
    }

    if ( !getenv( "LSAPI_PPID_NO_CHECK" ) )
        s_ppid = getppid();
    else
        ruby_unsetenv( "LSAPI_PPID_NO_CHECK" );
    if ( getenv( "RAILS_ROOT" )||getenv( "RAILS_ENV" ) )
        s_fn_add_env = add_env_rails;
        
    orig_stdin = rb_stdin;
    orig_stdout = rb_stdout;
#if RUBY_VERSION_CODE < 180
    orig_defout = rb_defout;
#endif
    orig_env = rb_const_get( rb_cObject, rb_intern("ENV") );

    env_copy = rb_funcall( orig_env, rb_intern( "to_hash" ), 0 );
    
    //tell the garbage collector it is a global variable, do not recycle it.
    rb_global_variable(&env_copy);
    
    //Do not need those environments after initialization
    remove_env = rb_str_new( "RAILS_ROOT", 10 );
    rb_funcall( env_copy, rb_intern( "delete" ), 1, remove_env );
    
    lsapi_env = rb_hash_new();
    //redefine ENV using a hash table, should be faster than char **environment
    rb_define_global_const("ENV", lsapi_env);
        
    LSAPI_Init();

    s_listenFd = g_req.m_fdListen;
    g_req.m_fdListen = -1;

    rb_define_global_function("eval_string_wrap", lsapi_eval_string_wrap, 1);

    cLSAPI = rb_define_class("LSAPI", rb_cObject);
    rb_define_singleton_method(cLSAPI, "accept", lsapi_s_accept, 0);

    rb_define_method(cLSAPI, "process", lsapi_process, 0 );
    //rb_define_method(cLSAPI, "initialize", lsapi_initialize, 0);
    rb_define_method(cLSAPI, "putc", lsapi_putc, 1);
    rb_define_method(cLSAPI, "write", lsapi_write, 1);
    rb_define_method(cLSAPI, "print", lsapi_print, -1);
    rb_define_method(cLSAPI, "printf", lsapi_printf, -1);
    rb_define_method(cLSAPI, "puts", lsapi_puts, -1);
    rb_define_method(cLSAPI, "<<", lsapi_addstr, 1);
    rb_define_method(cLSAPI, "flush", lsapi_flush, 0);
    rb_define_method(cLSAPI, "getc", lsapi_getc, 0);
    //rb_define_method(cLSAPI, "ungetc", lsapi_ungetc, 1);
    rb_define_method(cLSAPI, "gets", lsapi_gets, 0);
    rb_define_method(cLSAPI, "read", lsapi_read, -1);
    rb_define_method(cLSAPI, "eof", lsapi_eof, 0);
    rb_define_method(cLSAPI, "eof?", lsapi_eof, 0);
//    rb_define_method(cLSAPI, "close", lsapi_close, 0);
//    rb_define_method(cLSAPI, "closed?", lsapi_closed, 0);
    rb_define_method(cLSAPI, "binmode", lsapi_binmode, 0);
    //rb_define_method(cLSAPI, "isatty", lsapi_isatty, 0);
    //rb_define_method(cLSAPI, "tty?", lsapi_isatty, 0);
    rb_define_method(cLSAPI, "sync", lsapi_sync, 0);
    rb_define_method(cLSAPI, "sync=", lsapi_setsync, 1);
        
    s_req = Data_Make_Struct( cLSAPI, lsapi_data, lsapi_mark, free, s_req_data );
    s_req_data->req = &g_req;
    rb_stdin = rb_stdout = s_req;
#if RUBY_VERSION_CODE < 180
    rb_defout = s_req;
#endif
    rb_global_variable(&s_req );

    return;
}

